import { createParams } from "./core.js";
import { camelCaseToKebab, getValue, useMemoOptions } from "./utils.js";
import { createWrappers } from "./wrappers.js";
import { encodeParam, createListSerializer, defaultEncoder, defaultDecoder, stringSerializer, numberSerializer, booleanSerializer, datetimeSerializer, dateSerializer } from "./encoding.js";
const typeBasedOptions = {
  string: {
    ...stringSerializer
  },
  number: {
    ...numberSerializer
  },
  boolean: {
    ...booleanSerializer
  },
  datetime: {
    ...datetimeSerializer
  },
  date: {
    ...dateSerializer
  }
};
class InternalBuilder {
  constructor(state) {
    this.state = state;
  }
  withSerializer(serializer) {
    this.state = {
      ...this.state,
      ...serializer
    };
    return this;
  }
  validate(validator) {
    this.state = {
      ...this.state,
      ...validator
    };
    return this;
  }
  withDefault(value) {
    this.state = {
      ...this.state,
      defaultValue: value
    };
    return this;
  }
  asDialog() {
    return {
      state: {
        ...this.state,
        type: "dialog"
      }
    };
  }
}
const p = new Proxy({}, {
  get(_target, prop) {
    return (value) => {
      const key = prop;
      const options = typeBasedOptions[key] || {};
      const listOptions = key === "list" ? createListSerializer(value) : {};
      const defaultSerializers = {
        encode: defaultEncoder,
        decode: defaultDecoder
      };
      const state = {
        onError: () => {
        },
        validate: () => true,
        type: "value",
        defaultValue: null,
        updateType: "replaceIn",
        prefix: "",
        ...defaultSerializers,
        ...options,
        ...value || {},
        ...listOptions
      };
      return new InternalBuilder(state);
    };
  }
});
function create(schema, globalOptions) {
  const params = createParams();
  const wrappers = createWrappers(params);
  let internalGlobalOptions = globalOptions;
  const proxy = new Proxy({}, {
    get(_target, prop) {
      if (prop === "withOptions") {
        return (globalOptions2) => {
          internalGlobalOptions = {
            ...internalGlobalOptions,
            ...globalOptions2
          };
          return proxy;
        };
      }
      if (prop === "build") {
        return (values) => {
          const params2 = Object.entries(values).map(([key, value]) => {
            const schemaKey = camelCaseToKebab(key);
            const builder2 = schema[schemaKey];
            const { defaultValue, encode } = builder2.state;
            if (value === getValue(defaultValue)) {
              return void 0;
            }
            const encodedValue = encodeParam(value, encode);
            return `${key}=${encodedValue}`;
          }).filter(Boolean);
          return params2.join("&");
        };
      }
      if (prop === "batch") {
        return (fn) => {
          const { paramsStore, api } = params.getInternals();
          paramsStore.batch(() => {
            api.batch(fn);
          });
        };
      }
      const schemaField = camelCaseToKebab(prop);
      const builder = schema[schemaField];
      const options = builder.state;
      const simpleMap = () => {
        const typedOptions = options;
        return {
          useSet: (options2) => {
            const memoedOptions = useMemoOptions(typedOptions, internalGlobalOptions, options2);
            return params.useParamSet(prop, memoedOptions);
          },
          use: (options2) => {
            const memoedOptions = useMemoOptions(typedOptions, internalGlobalOptions, options2);
            return params.useParam(prop, memoedOptions);
          }
        };
      };
      if (options.type === "value") {
        return simpleMap();
      } else if (options.type === "dialog") {
        const typedOptionsWithDialogDefault = {
          ...options,
          defaultValue: {
            isOpen: false,
            state: getValue(options.defaultValue)
          }
        };
        return {
          //todo: better handle default options
          useSet: (options2) => {
            const memoedOptions = useMemoOptions(typedOptionsWithDialogDefault, internalGlobalOptions, options2);
            return wrappers.useDialogParamSet(prop, memoedOptions);
          },
          use: (options2) => {
            const memoedOptions = useMemoOptions(typedOptionsWithDialogDefault, internalGlobalOptions, options2);
            return wrappers.useDialogParam(prop, memoedOptions);
          }
        };
      }
      throw new Error("Invalid type");
    }
  });
  return proxy;
}
export {
  create,
  p
};
//# sourceMappingURL=params.js.map
